#!/usr/bin/env python3
"""
GPGME Wrapper Wrapper
by Javantea
March 24-25, 2017

For BikeIM Mail1 Mail Client
"""
from __future__ import print_function
import gpg
import sys

if sys.hexversion >= 0x3000000:
	baseUnicode = str
else:
	baseUnicode = unicode
#end if

initted = False

# Fixing a bug in gpg.errors.
gpg.errors.EOF = 0x3fff

"""
def keylist(ctx, pattern, secret=False):
	return keylist_all(ctx, pattern, secret)
#end def keylist(ctx, pattern, secret=False)

# A copy of gpg/core.py:Context.op_keylist_all()
# Fixes the EOF issue
def keylist_all(ctx, *args, **kwargs):
	ctx.op_keylist_start(*args, **kwargs)
	key = ctx.op_keylist_next()
	while key:
		yield key
		try:
			key = ctx.op_keylist_next()
		except gpg.errors.GPGMEError as excp:
			key = None
			#if excp.getcode() != gpg.errors.EOF:
				#raise excp
				#print('exception', excp.getcode(), gpg.errors.EOF)
	ctx.op_keylist_end()
#end def keylist_all(ctx, *args, **kwargs)
"""

def GPG_init():
	global initted
	if initted: return
	initted = True
	proto = gpg.constants.protocol.OpenPGP
	gpg.core.engine_check_version(proto)
#end def GPG_init()

def GPG_decrypt(ciphertext):
	"""
	Decrypt an OpenPGP ciphertext.
	Useful always.
	To check the result, first check if it's None.
	Then split into verify_result and return data.
	If return data is None, it's an error.
	If verify_result.signatures == [], it's unsigned (bad news.)
	Then you have to check verify_result.signatures[i].fpr actually comes from 
	the person you expect.
	"""
	GPG_init()
	c = gpg.Context()
	source = gpg.Data(ciphertext)
	sink = gpg.Data()
	try:
		c.op_decrypt_verify(source, sink)
	except gpg.errors.GPGMEError as err:
		print("Error: decrypt failed with", err)
		return None
	#end try
	result = c.op_decrypt_result()
	if not result:
		return result, None
	#end if
	# NOTE: result is thrown away and has: 'file_name', 'recipients', 'session_key', 'unsupported_algorithm', 'wrong_key_usage'
	# Timestamp would be useful but only an encrypted or macced value would be worthwhile (since an attacker can change an un macced unencrypted timestamp).
	#print('q', dir(result))
	verify_result = c.op_verify_result()
	if not verify_result:
		return verify_result, None
	#end if
	#sink.seek(0, 0)
	#print('wtfh?', sink.read())
	sink.seek(0, 0)
	return verify_result, sink.read()
#end def GPG_decrypt(ciphertext)

def GPG_sign(plaintext, keyid=None, detach=True):
	"""
	Sign a piece of data in detached mode.
	Useful for software or pgp/mime mail.
	plaintext is bytes you want signed.
	keyid is the key to use to sign. Default is the default key.
	"""
	GPG_init()
	c = gpg.Context()
	#c.set_textmode(True)
	c.set_armor(True)
	if keyid:
		c.signers_clear()
		c.signers_add(c.get_key(keyid))
	#end if
	# Never let a non-bytes past because gpg.Data breaks with baseUnicode
	if isinstance(plaintext, baseUnicode):
		plaintext = plaintext.encode('utf-8', 'surrogateescape')
	#end if
	source = gpg.Data(plaintext)
	sink = gpg.Data()
	try:
		c.op_sign(source, sink, detach and gpg.constants.sig.mode.DETACH)
	except gpg.errors.GPGMEError as ge:
		#print('s e', dir(ge), ge.error, gpg.errors.gpgme.gpgme_strerror(ge.error))
		if gpg.errors.gpgme.gpgme_strerror(ge.error) == 'Unusable secret key':
			return ge, None
		#end if
		# Don't let any error past
		raise ge
	#end try
	result = c.op_sign_result()
	if not result:
		return result, None
	#end if
	if len(result.signatures) != 1:
		return result, None
	#end if
	# If more validation is necessary let me know.
	sink.seek(0, 0)
	return result, sink.read()
#end def GPG_sign(plaintext, keyid=None, detach=True)

def GPG_clearsign(plaintext, keyid=None):
	"""
	Sign a piece of data in clearsign mode.
	Useful for websites, checksums, or traditional pgp mail.
	plaintext is bytes you want signed.
	keyid is the key to use to sign. Default is the default key.
	"""
	GPG_init()
	c = gpg.Context()
	c.set_armor(True)
	if keyid:
		c.signers_clear()
		c.signers_add(c.get_key(keyid))
	#end if
	# Never let a non-bytes past because gpg.Data breaks with baseUnicode
	if isinstance(plaintext, baseUnicode):
		plaintext = plaintext.encode('utf-8', 'surrogateescape')
	#end if
	source = gpg.Data(plaintext)
	sink = gpg.Data()
	try:
		c.op_sign(source, sink, gpg.constants.sig.mode.CLEAR)
	except gpg.errors.GPGMEError as ge:
		#print('s e', dir(ge), ge.error, gpg.errors.gpgme.gpgme_strerror(ge.error))
		if gpg.errors.gpgme.gpgme_strerror(ge.error) == 'Unusable secret key':
			return ge, None
		#end if
		# Don't let any error past
		raise ge
	#end try
	result = c.op_sign_result()
	if not result:
		return result, None
	#end if
	if len(result.signatures) != 1:
		return result, None
	#end if
	# If more validation is necessary let me know.
	sink.seek(0, 0)
	return result, sink.read()
#end def GPG_clearsign(plaintext, keyid)

def GPG_encrypt(plaintext, recipient):
	"""
	Encrypt a piece of data with a list of public keys.
	plaintext is bytes you want signed.
	Does not sign the data, you want GPG_sign_encrypt.
	Not signing encrypted data is like tapping morse code on a wall and wishing 
	that it gets to its destination. Yes, the upstream does have consequences 
	for replacing your message with their own, but what consequences do they 
	face?
	This function is used in pgp/mime because signing occurs before encryption 
	for an odd reason.
	"""
	GPG_init()
	c = gpg.Context()
	c.set_armor(True)
	# Never let a non-bytes past because gpg.Data breaks with baseUnicode
	if isinstance(plaintext, baseUnicode):
		plaintext = plaintext.encode('utf-8', 'surrogateescape')
	#end if
	source = gpg.Data(plaintext)
	sink = gpg.Data()
	keys = []
	for r in recipient:
		r_keys = list(c.keylist(r))
		if len(r_keys) == 1:
			keys.append(r_keys[0])
		else:
			return (False, (r, r_keys))
		#end if
	#next r
	# FIXME: Always trust is not our method.
	c.op_encrypt(keys, gpg.constants.ENCRYPT_ALWAYS_TRUST, source, sink)
	result = c.op_encrypt_result()
	if not result:
		return result, None
	#end if
	if result.invalid_recipients:
		print("Error: Invalid recipients.")
		return result, None
	#end if
	sink.seek(0, 0)
	return result, sink.read()
#end def GPG_encrypt(plaintext, recipient)

def GPG_sign_encrypt(plaintext, recipient, keyid=None):
	"""
	Sign and Encrypt a piece of data with a list of public keys.
	plaintext is bytes you want signed.
	keyid is the key to use to sign. Default is the default key.
	"""
	GPG_init()
	c = gpg.Context()
	c.set_armor(True)
	if keyid:
		c.signers_clear()
		c.signers_add(c.get_key(keyid))
	#end if
	# Never let a non-bytes past because gpg.Data breaks with baseUnicode
	if isinstance(plaintext, baseUnicode):
		plaintext = plaintext.encode('utf-8', 'surrogateescape')
	#end if
	source = gpg.Data(plaintext)
	sink = gpg.Data()
	keys = []
	for r in recipient:
		r_keys = list(c.keylist(r))
		if len(r_keys) == 1:
			keys.append(r_keys[0])
		else:
			return (False, (r, r_keys))
		#end if
	#next r
	# FIXME: Always trust is not our method.
	try:
		c.op_encrypt_sign(keys, gpg.constants.ENCRYPT_ALWAYS_TRUST, source, sink)
	except gpg.errors.GPGMEError as ge:
		#print('s e', dir(ge), ge.error, gpg.gpgme.gpgme_strerror(ge.error))
		if gpg.errors.gpgme.gpgme_strerror(ge.error) == 'Unusable secret key':
			return ge, None
		#end if
		# Don't let any error past
		raise ge
	#end try
	result = c.op_encrypt_result()
	if result.invalid_recipients:
		print("Error: Invalid recipients.")
		return result, None
	#end if
	result = c.op_sign_result()
	if not result:
		print("Error: Signature failed.")
		return result, None
	#end if
	if len(result.signatures) != 1:
		return result, None
	#end if
	# If more validation is necessary let me know.
	sink.seek(0, 0)
	return result, sink.read()
#end def GPG_sign_encrypt(plaintext, recipient, keyid=None)

def GPG_verify(clearsigned):
	"""
	Verify a clearsigned message.
	Useful for traditional pgp email.
	"""
	GPG_init()
	c = gpg.Context()
	c.set_armor(True)
	c.set_textmode(True)
	source = gpg.Data(clearsigned)
	sink = gpg.Data()
	try:
		c.op_verify(source, None, sink)
	except gpg.errors.GPGMEError as ge:
		# This occurs when a person replies to a traditional signed e-mail without stripping the sig.
		# Or when a person types -----BEGIN PGP SIGNED MESSAGE----- into an unsigned message.
		print("Warning: Verification error", ge)
		return None, None
	#end try
	result = c.op_verify_result()
	if not result:
		return result, None
	#end if
	sink.seek(0, 0)
	return result, sink.read()
#end def GPG_verify(clearsigned)

def GPG_verify2(data, sig):
	"""
	Verify a detached signed message.
	Useful for software or pgp/mime email.
	"""
	GPG_init()
	c = gpg.Context()
	c.set_armor(True)
	#c.set_textmode(True)
	dd = data.replace(b'\n',b'\r\n')
	#print(data, dd, sig)
	source = gpg.Data(dd)
	sig_data = gpg.Data(sig)
	#sink = gpg.Data()
	c.op_verify(sig_data, source, None)
	result = c.op_verify_result()
	if not result:
		return result
	#end if
	return result
#end def GPG_verify2(clearsigned)

def GPG_import(key):
	"""
	Import a public key
	
	"""
	GPG_init()
	if isinstance(key, baseUnicode):
		key = key.encode('utf-8', 'surrogateescape')
	#end if
	source = gpg.Data(key)
	c = gpg.Context()
	c.op_import(source)
	result = c.op_import_result()
	return result
#end def GPG_import(key)

def GPG_genkey(name, email, comment='', expire=0):
	"""
	Generate a keypair
	
	"""
	GPG_init()
	config = """<GnupgKeyParms format="internal">
Key-Type: RSA
Key-Length: 4096
Subkey-Type: RSA
Subkey-Length: 4096
Name-Real: {0}
Name-Email: {1}
{2}Expire-Date: {3}
</GnupgKeyParms>""".format(name, email, (comment and 'Name-Comment: {2}\n') or '', expire)
	#print(repr(config))
	c = gpg.Context()
	try:
		c.op_genkey(config, None, None)
	except gpg.errors.GPGMEError as ge:
		return ge
	#end try
	result = c.op_genkey_result()
	return result
#end def GPG_genkey(name, email, expire)

def getKeylist(recipient):
	"""
	Find the key/keys for a certain keyid or e-mail address/name
	"""
	GPG_init()
	c = gpg.Context()
	c.set_armor(True)
	return list(c.keylist(recipient))
#end def getKeylist(recipient)

def summary_to_string(summary):
	strsummary = [(x, getattr(gpg.constants.sigsum, x)) for x in dir(gpg.constants.sigsum) if x in ['BAD_POLICY', 'CRL_MISSING', 'CRL_TOO_OLD', 'GREEN', 'KEY_EXPIRED', 'KEY_MISSING', 'KEY_REVOKED', 'RED', 'SIG_EXPIRED', 'SYS_ERROR', 'TOFU_CONFLICT', 'VALID']]
	message = ' '.join([ssum for ssum, sum_i in strsummary if summary & sum_i])
	return message
#end def 


def main():
	import time
	ciphertext1 = '-----BEGIN PGP MESSAGE-----\nVersion: GnuPG v2.0.20 (GNU/Linux)\n\nhQEOAzhXk43XKEchEAQAzk+HHS4v33tV5v8ReOusMmeyZyK/c/98MUiblXPQYyyI\nMdNXMXieB77LzwJiD+tglGyL9dMcnFFpMq4jDHykQoJoUa61WeOXF8Fv0Sz0Fm2o\ns55dKEEXBK4X/lvs/TME9UlIvJUBXj0txX93n3x6qh3gO+mQArCixH5p31CShfID\n/0ZbXZGcx7vdBSJhN2i4E26BhWgpjr+XygQVnhtLpzdk3AXHjq0+bB5IgzZAfPct\ngetiQSaRyAulU8hyThPVcnRJxuBs/X2KyEBjJZvSY5ordkLdcYJSYXNQNKLzP+EN\nMUJRMARINASwLhiscdXfzwA60/6CelVJ5pSRvg8hDlwv0pQB05Zasspb5WlCDSZ0\nmLqadlReCLOmlz12EUJwFWXrUnYZtEuf2YRpUOuG0XMlLtw1UjKc1XFiDgsSCvpX\nRlNndDCodVP1w3LUEP6yh1wRjQBuPy7MzuL3uKUisJpgu5mFGJjAJ9iDCb4ynjMh\nZ56g44LrzRKBcyOMF619BFUlhbip/E14H53iP0BCjJ6BftJMNx1e\n=T4KQ\n-----END PGP MESSAGE-----\n'
	result = GPG_decrypt(ciphertext1)
	if result == None or len(result) != 2:
		print("Error: GPG_decrypt result is invalid", result)
		return
	#end if
	result, plaintext = result
	good_sig, fpr, created = False, None, 0
	if len(result.signatures) > 0:
		good_sig = (result.signatures[0].summary & gpg.constants.sigsum.VALID)
		fpr = result.signatures[0].fpr
		created = result.signatures[0].timestamp
		if result.signatures[0].key:
			name = str(result.signatures[0].key)
		else:
			keys = getKeylist(fpr)
			if len(keys) == 1:
				name = ', '.join(['{0} <{1}>'.format(uid.name, uid.email) for uid in keys[0].uids])
			elif len(keys) > 1:
				name = 'Multiple keys: '
				for key in keys:
					name += ', '.join(['{0} <{1}>'.format(uid.name, uid.email) for uid in key.uids]) + '; '
				#next key
			else:
				name = 'Unknown ' + fpr
			#end if
		#end if
	#end if
	if good_sig:
		good_msg = 'Good signature from {0}'.format(name)
		sig = ('mail1: Signature made {0}\nPrimary key fingerprint: ' + fpr).format(time.strftime('%a %d %b %Y %T %Z', time.localtime(float(created))))
	else:
		message = ''
		if result.signatures:
			message = summary_to_string(result.signatures[0].summary)
		#end if
		print("Warning: Bad signature 1", repr(result), result.signatures, message)
		good_msg = 'Signed by {0}'.format(name)
		sig = "Warning: Bad signature " + message
	#end if
	if not good_sig:
		if len(result.signatures) == 0 or (result.signatures[0].summary & gpg.constants.sigsum.KEY_EXPIRED == 0):
			raise Exception("Error: Bad signature")
		#end if
		print('Not green')
	#end if
	print(repr(sig))
	print(repr(good_msg))
	print(repr(plaintext))
#end def main()

if __name__ == '__main__':
	main()
#end if
